@@ -7,221 +7,12 @@ from traitlets import (
     default, validate, TraitError
 )
 
-from traitlets.utils.bunch import Bunch
+from .basemaps import basemaps
 
 from ._version import EXTENSION_VERSION
 
 def_loc = [0.0, 0.0]
 
-
-basemaps = Bunch(
-    OpenStreetMap = Bunch(
-        Mapnik = dict(
-            url = 'https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
-            max_zoom = 19,
-            attribution = 'Map data (c) <a href="https://openstreetmap.org">OpenStreetMap</a> contributors',
-            name = 'OpenStreetMap.Mapnik'
-        ),
-        BlackAndWhite = dict(
-            url = 'http://{s}.tiles.wmflabs.org/bw-mapnik/{z}/{x}/{y}.png',
-            max_zoom = 18,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'OpenStreetMap.BlackAndWhite',
-        ),
-        DE = dict(
-            url = 'http://{s}.tile.openstreetmap.de/tiles/osmde/{z}/{x}/{y}.png',
-            max_zoom = 18,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'OpenStreetMap.DE'
-        ),
-        France = dict(
-            url = 'http://{s}.tile.openstreetmap.fr/osmfr/{z}/{x}/{y}.png',
-            max_zoom = 20,
-            attribution = '&copy; Openstreetmap France | &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'OpenStreetMap.France'
-        ),
-        HOT = dict(
-            url = 'http://{s}.tile.openstreetmap.fr/hot/{z}/{x}/{y}.png',
-            max_zoom = 19,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>, Tiles courtesy of <a href="http://hot.openstreetmap.org/" target="_blank">Humanitarian OpenStreetMap Team</a>',
-            name = 'OpenStreetMap.HOT'
-       )
-    ),
-    OpenTopoMap = dict(
-        url = 'https://{s}.tile.opentopomap.org/{z}/{x}/{y}.png',
-        max_zoom = 17,
-        attribution = 'Map data: &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>, <a href="http://viewfinderpanoramas.org">SRTM</a> | Map style: &copy; <a href="https://opentopomap.org">OpenTopoMap</a> (<a href="https://creativecommons.org/licenses/by-sa/3.0/">CC-BY-SA</a>)',
-        name = 'OpenTopMap'
-    ),
-    OpenMapSurfer = Bunch(
-        Roads = dict(
-            url = 'http://korona.geog.uni-heidelberg.de/tiles/roads/x={x}&y={y}&z={z}',
-            max_zoom = 20,
-            attribution = 'Imagery from <a href="http://giscience.uni-hd.de/">GIScience Research Group @ University of Heidelberg</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'OpenMapSurfer.Roads'
-        ),
-        Grayscale = dict(
-            url = 'http://korona.geog.uni-heidelberg.de/tiles/roadsg/x={x}&y={y}&z={z}',
-            max_zoom = 19,
-            attribution = 'Imagery from <a href="http://giscience.uni-hd.de/">GIScience Research Group @ University of Heidelberg</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'OpenMapSurfer.Grayscale'
-        )
-    ),
-    Hydda = Bunch(
-        Full = dict(
-            url = 'http://{s}.tile.openstreetmap.se/hydda/full/{z}/{x}/{y}.png',
-            max_zoom = 18,
-            attribution = 'Tiles courtesy of <a href="http://openstreetmap.se/" target="_blank">OpenStreetMap Sweden</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'Hydda.Full'
-        ),
-        Base = dict(
-            url = 'http://{s}.tile.openstreetmap.se/hydda/base/{z}/{x}/{y}.png',
-            max_zoom = 18,
-            attribution = 'Tiles courtesy of <a href="http://openstreetmap.se/" target="_blank">OpenStreetMap Sweden</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'Hydda.Base'
-        ),
-    ),
-    Esri = Bunch(
-        WorldStreetMap = dict(
-            url = 'http://server.arcgisonline.com/ArcGIS/rest/services/World_Street_Map/MapServer/tile/{z}/{y}/{x}',
-            max_zoom = 20,
-            attribution = 'Tiles &copy; Esri &mdash; Source: Esri, DeLorme, NAVTEQ, USGS, Intermap, iPC, NRCAN, Esri Japan, METI, Esri China (Hong Kong), Esri (Thailand), TomTom, 2012',
-            name = 'Esri.WoldStreetMap'
-        ),
-        DeLorme = dict(
-            url = 'http://server.arcgisonline.com/ArcGIS/rest/services/Specialty/DeLorme_World_Base_Map/MapServer/tile/{z}/{y}/{x}',
-            min_zoom = 1,
-            max_zoom = 11,
-            attribution = 'Tiles &copy; Esri &mdash; Copyright: &copy;2012 DeLorme',
-            name = 'Esri.DeLorme'
-        ),
-        WorldTopoMap = dict(
-            url = 'http://server.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/{z}/{y}/{x}',
-            max_zoom = 20,
-            attribution = 'Tiles &copy; Esri &mdash; Esri, DeLorme, NAVTEQ, TomTom, Intermap, iPC, USGS, FAO, NPS, NRCAN, GeoBase, Kadaster NL, Ordnance Survey, Esri Japan, METI, Esri China (Hong Kong), and the GIS User Community',
-            name = 'Esri.WorldTopoMap'
-        ),
-        WorldImagery = dict(
-            url = 'http://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
-            max_zoom = 20,
-            attribution = 'Tiles &copy; Esri &mdash; Source: Esri, i-cubed, USDA, USGS, AEX, GeoEye, Getmapping, Aerogrid, IGN, IGP, UPR-EGP, and the GIS User Community',
-            name = 'Esri.WorldImagery'
-        ),
-        NatGeoWorldMap = dict(
-            url = 'http://server.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/{z}/{y}/{x}',
-            max_zoom = 16,
-            attribution = 'Tiles &copy; Esri &mdash; National Geographic, Esri, DeLorme, NAVTEQ, UNEP-WCMC, USGS, NASA, ESA, METI, NRCAN, GEBCO, NOAA, iPC',
-            name = 'Esri.NatGeoWorldMap'
-        ),
-    ),
-    HikeBike = Bunch(
-        HikeBike = dict(
-            url = 'http://{s}.tiles.wmflabs.org/hikebike/{z}/{x}/{y}.png',
-            max_zoom = 19,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'HikeBike.HikeBike'
-        ),
-    ),
-    Stamen = Bunch(
-        Terrain = dict(
-            url = 'http://stamen-tiles-a.a.ssl.fastly.net/terrain/{z}/{x}/{y}.png',
-            max_zoom = 18,
-            attribution = 'Map tiles by <a href="http://stamen.com">Stamen Design</a>, <a href="http://creativecommons.org/licenses/by/3.0">CC BY 3.0</a> &mdash; Map data &copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>',
-            name = 'Stamen.Terrain'
-        ),
-    ),
-    MtbMap = dict(
-        url = 'http://tile.mtbmap.cz/mtbmap_tiles/{z}/{x}/{y}.png',
-        max_zoom = 20,
-        attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &amp; USGS',
-        name = 'MtbMap'
-    ),
-    CartoDB = Bunch(
-        Positron = dict(
-            url = 'http://tile.mtbmap.cz/mtbmap_tiles/{z}/{x}/{y}.png',
-            max_zoom = 20,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &amp; USGS',
-            name = 'CartoDB.Positron'
-        ),
-        DarkMatter = dict(
-            url = 'http://c.basemaps.cartocdn.com/dark_all/{z}/{x}/{y}.png',
-            max_zoom = 19,
-            attribution = '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a> &copy; <a href="http://cartodb.com/attributions">CartoDB</a>',
-            name = 'CartoDB.DarkMatter'
-        )
-    ),
-    NASAGIBS = Bunch(
-        ModisTerraTrueColorCR = dict(
-            url = 'https://map1.vis.earthdata.nasa.gov/wmts-webmerc/MODIS_Terra_CorrectedReflectance_TrueColor/default/%s/GoogleMapsCompatible_Level9/{z}/{y}/{x}.jpg',
-            max_zoom = 9,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'NASAGIBS.ModisTerraTrueColorCR'
-        ),
-        ModisTerraBands367CR = dict(
-            url = 'https://map1.vis.earthdata.nasa.gov/wmts-webmerc/MODIS_Terra_CorrectedReflectance_Bands367/default/%s/GoogleMapsCompatible_Level9/{z}/{y}/{x}.jpg',
-            max_zoom = 9,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'NASAGIBS.ModisTerraBands367CR'
-        ),
-        ModisTerraBands721CR = dict(
-            url = 'https://map1.vis.earthdata.nasa.gov/wmts-webmerc/MODIS_Terra_CorrectedReflectance_Bands721/default/%s/GoogleMapsCompatible_Level9/{z}/{y}/{x}.jpg',
-            max_zoom = 9,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'NASAGIBS.MidsTerraBands721CR'
-        ),
-        ModisAquaTrueColorCR = dict(
-            url = 'https://map1.vis.earthdata.nasa.gov/wmts-webmerc/MODIS_Aqua_CorrectedReflectance_TrueColor/default/%s/GoogleMapsCompatible_Level9/{z}/{y}/{x}.jpg',
-            max_zoom = 9,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'NASAGIBS.ModisAquaTrueColorCR'
-        ),
-        ModisAquaBands721CR = dict(
-            url = 'https://map1.vis.earthdata.nasa.gov/wmts-webmerc/MODIS_Aqua_CorrectedReflectance_Bands721/default/%s/GoogleMapsCompatible_Level9/{z}/{y}/{x}.jpg',
-            max_zoom = 9,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'NASAGIBS.ModisAquaBands721CR'
-        ),
-        ViirsEarthAtNight2012 = dict(
-            url = 'http://map1.vis.earthdata.nasa.gov/wmts-webmerc/VIIRS_CityLights_2012/default/2012-08-01/GoogleMapsCompatible_Level8/{z}/{y}/{x}.jpg',
-            max_zoom = 8,
-            attribution = 'Imagery provided by services from the Global Imagery Browse Services (GIBS), operated by the NASA/GSFC/Earth Science Data and Information System (<a href="https://earthdata.nasa.gov">ESDIS</a>) with funding provided by NASA/HQ.',
-            name = 'BASAGIBS.ViirsEarthAtNight2012'
-        )
-    ),
-    Strava = Bunch(
-        All = dict(
-            url = 'https://heatmap-external-a.strava.com//tiles/all/hot/{z}/{x}/{y}.png?v=19',
-            max_zoom = 15,
-            attribution = 'href="https://labs.strava.com/heatmap">Strava 2017</a>',
-            name = 'Strava.All'
-        ),
-        Ride = dict(
-            url = 'https://heatmap-external-a.strava.com//tiles/ride/hot/{z}/{x}/{y}.png?v=19',
-            max_zoom = 15,
-            attribution = 'href="https://labs.strava.com/heatmap">Strava 2017</a>',
-            name = 'Strava.Ride'
-        ),
-        Run = dict(
-            url = 'https://heatmap-external-a.strava.com//tiles/run/bluered/{z}/{x}/{y}.png?v=19',
-            max_zoom = 15,
-            attribution = 'href="https://labs.strava.com/heatmap">Strava 2017</a>',
-            name = 'Strava.Run'
-        ),
-        Water = dict(
-            url = 'https://heatmap-external-a.strava.com//tiles/water/blue/{z}/{x}/{y}.png?v=19',
-            max_zoom = 15,
-            attribution = 'href="https://labs.strava.com/heatmap">Strava 2017</a>',
-            name = 'Strava.Water'
-        ),
-        Winter = dict(
-            url = 'https://heatmap-external-a.strava.com//tiles/winter/hot/{z}/{x}/{y}.png?v=19',
-            max_zoom = 15,
-            attribution = 'href="https://labs.strava.com/heatmap">Strava 2017</a>',
-            name = 'Strava.Winter'
-        )
-    )
-)
-
 def basemap_to_tiles(bm, day='yesterday', **kwargs):
     # Format the URL with modisdate
     from datetime import date, timedelta
@@ -240,6 +31,7 @@ def basemap_to_tiles(bm, day='yesterday', **kwargs):
         **kwargs
     )
 
+
 class LayerException(TraitError):
     pass
 
@@ -273,6 +65,8 @@ class Layer(Widget, InteractMixin):
     bottom = Bool().tag(sync=True)
     options = List(trait=Unicode).tag(sync=True)
 
+    popup = Instance(Widget, allow_none=True, default_value=None).tag(sync=True, **widget_serialization)
+
     @default('options')
     def _default_options(self):
         return [name for name in self.traits(o=True)]
@@ -316,10 +110,35 @@ class Marker(UILayer):
     def on_move(self, callback, remove=False):
         self._move_callbacks.register_callback(callback, remove=remove)
 
+
 class Popup(UILayer):
     _view_name = Unicode('LeafletPopupView').tag(sync=True)
     _model_name = Unicode('LeafletPopupModel').tag(sync=True)
 
+    location = List(def_loc).tag(sync=True)
+    child = Instance(
+        DOMWidget, allow_none=True, default_value=None
+    ).tag(sync=True, **widget_serialization)
+
+    max_width = Int(300).tag(sync=True, o=True)
+    min_width = Int(50).tag(sync=True, o=True)
+    max_height = Int(allow_none=True, default_value=None).tag(sync=True, o=True)
+    auto_pan = Bool(True).tag(sync=True, o=True)
+    auto_pan_padding_top_left = List(allow_none=True, default_value=None).tag(sync=True, o=True)
+    auto_pan_padding_bottom_right = List(allow_none=True, default_value=None).tag(sync=True, o=True)
+    auto_pan_padding = List([5, 5]).tag(sync=True, o=True)
+    keep_in_view = Bool(False).tag(sync=True, o=True)
+    close_button = Bool(True).tag(sync=True, o=True)
+    auto_close = Bool(True).tag(sync=True, o=True)
+    close_on_escape_key = Bool(True).tag(sync=True, o=True)
+    class_name = Unicode('').tag(sync=True, o=True)
+
+    options = List(trait=Unicode).tag(sync=True)
+
+    @default('options')
+    def _default_options(self):
+        return [name for name in self.traits(o=True)]
+
 
 class RasterLayer(Layer):
     _view_name = Unicode('LeafletRasterLayerView').tag(sync=True)
@@ -432,19 +251,19 @@ class Rectangle(Polygon):
     bounds = List(help="list of SW and NE location tuples").tag(sync=True)
 
 
-class Circle(Path):
-    _view_name = Unicode('LeafletCircleView').tag(sync=True)
-    _model_name = Unicode('LeafletCircleModel').tag(sync=True)
-
-    location = List(def_loc).tag(sync=True)
-    radius = Int(1000, help="radius of circle in meters").tag(sync=True)
-
-
-class CircleMarker(Circle):
+class CircleMarker(Path):
     _view_name = Unicode('LeafletCircleMarkerView').tag(sync=True)
     _model_name = Unicode('LeafletCircleMarkerModel').tag(sync=True)
 
-    radius = Int(10, help="radius of circle in pixels").tag(sync=True)
+    location = List(def_loc).tag(sync=True)
+    radius = Int(10, help="radius of circle in pixels").tag(sync=True, o=True)
+
+
+class Circle(CircleMarker):
+    _view_name = Unicode('LeafletCircleView').tag(sync=True)
+    _model_name = Unicode('LeafletCircleModel').tag(sync=True)
+
+    radius = Int(1000, help="radius of circle in meters").tag(sync=True, o=True)
 
 
 class MarkerCluster(Layer):
@@ -458,7 +277,36 @@ class LayerGroup(Layer):
     _view_name = Unicode('LeafletLayerGroupView').tag(sync=True)
     _model_name = Unicode('LeafletLayerGroupModel').tag(sync=True)
 
-    layers = List(Instance(Layer)).tag(sync=True, **widget_serialization)
+    layers = Tuple(trait=Instance(Layer)).tag(sync=True, **widget_serialization)
+
+    layer_ids = List()
+
+    @validate('layers')
+    def _validate_layers(self, proposal):
+        """Validate layers list.
+
+        Makes sure only one instance of any given layer can exist in the
+        layers list.
+        """
+        self.layer_ids = [l.model_id for l in proposal.value]
+        if len(set(self.layer_ids)) != len(self.layer_ids):
+            raise LayerException('duplicate layer detected, only use each layer once')
+        return proposal.value
+
+    def add_layer(self, layer):
+        if isinstance(layer, dict):
+            layer = basemap_to_tiles(layer)
+        if layer.model_id in self.layer_ids:
+            raise LayerException('layer already in layergroup: %r' % layer)
+        self.layers = tuple([l for l in self.layers] + [layer])
+
+    def remove_layer(self, layer):
+        if layer.model_id not in self.layer_ids:
+            raise LayerException('layer not on in layergroup: %r' % layer)
+        self.layers = tuple([l for l in self.layers if l.model_id != layer.model_id])
+
+    def clear_layers(self):
+        self.layers = ()
 
 
 class FeatureGroup(Layer):
@@ -500,11 +348,6 @@ class GeoJSON(FeatureGroup):
         self._hover_callbacks.register_callback(callback, remove=remove)
 
 
-class MultiPolyline(FeatureGroup):
-    _view_name = Unicode('LeafletMultiPolylineView').tag(sync=True)
-    _model_name = Unicode('LeafletMultiPolylineModel').tag(sync=True)
-
-
 class MultiPolygon(FeatureGroup):
     _view_name = Unicode('LeafletMultiPolygonView').tag(sync=True)
     _model_name = Unicode('LeafletMultiPolygonModel').tag(sync=True)
@@ -538,17 +381,16 @@ class LayersControl(Control):
 class SplitMapControl(Control):
     _view_name = Unicode('LeafletSplitMapControlView').tag(sync=True)
     _model_name = Unicode('LeafletSplitMapControlModel').tag(sync=True)
-    _draw_callbacks = Instance(CallbackDispatcher, ())
 
-    leftLayer = Instance(TileLayer).tag(sync=True, **widget_serialization)
-    rightLayer = Instance(TileLayer).tag(sync=True, **widget_serialization)
+    left_layer = Instance(TileLayer).tag(sync=True, **widget_serialization)
+    right_layer = Instance(TileLayer).tag(sync=True, **widget_serialization)
 
-    @default('leftLayer')
-    def _default_leftLayer(self):
+    @default('left_layer')
+    def _default_left_layer(self):
         return TileLayer()
 
-    @default('rightLayer')
-    def _default_rightLayer(self):
+    @default('right_layer')
+    def _default_right_layer(self):
         return TileLayer()
 
     def __init__(self, **kwargs):
@@ -730,7 +572,7 @@ class Map(DOMWidget, InteractMixin):
         if isinstance(new, dict):
             new = basemap_to_tiles(new)
         if old.model_id not in self.layer_ids:
-            raise LayerException('layer not on map: %r' % layer)
+            raise LayerException('Could not substitute layer: layer not on map.')
         self.layers = tuple([new if l.model_id == old.model_id else l for l in self.layers])
 
     def clear_layers(self):
@@ -786,17 +628,11 @@ class Map(DOMWidget, InteractMixin):
         return self
 
     # Event handling
-    _moveend_callbacks = Instance(CallbackDispatcher, ())
     _interaction_callbacks = Instance(CallbackDispatcher, ())
 
     def _handle_leaflet_event(self, _, content, buffers):
-        if content.get('event', '') == 'moveend':
-            self._moveend_callbacks(**content)
         if content.get('event', '') == 'interaction':
             self._interaction_callbacks(**content)
 
-    def on_moveend(self, callback, remove=False):
-        self._moveend_callbacks.register_callback(callback, remove=remove)
-
     def on_interaction(self, callback, remove=False):
         self._interaction_callbacks.register_callback(callback, remove=remove)
